from typing import Callable, Union, Type, Any, Optional, Dict, TYPE_CHECKING, TypeVar
from uuid import uuid4, UUID

from unipipeline.answer.uni_answer_message import UniAnswerMessage
from unipipeline.message.uni_message import UniMessage

if TYPE_CHECKING:
    from unipipeline.worker.uni_worker import UniWorker


TInputMessage = TypeVar('TInputMessage', bound=UniMessage)
TAnswMessage = TypeVar('TAnswMessage', bound=UniMessage)


class UniWorkerConsumerManager:
    def __init__(self, send: Callable[[Union[Type['UniWorker[Any, Any]'], str], Union[Dict[str, Any], UniMessage], bool, bool], Optional[UniAnswerMessage[UniMessage]]]) -> None:
        self._send = send
        self._id = uuid4()

    @property
    def id(self) -> UUID:
        return self._id

    def stop_consuming(self) -> None:
        raise NotImplementedError(f'{type(self).__name__}.stop_consuming was not implemented')  # TODO

    def exit(self) -> None:
        raise NotImplementedError(f'{type(self).__name__}.exit was not implemented')  # TODO

    def get_answer_from(self, worker: Union[Type['UniWorker[TInputMessage, TAnswMessage]'], str], data: Union[Dict[str, Any], TInputMessage]) -> UniAnswerMessage[TAnswMessage]:
        return self._send(worker, data, False, True)  # type: ignore

    def send_to(self, worker: Union[Type['UniWorker[Any, Any]'], str], data: Union[Dict[str, Any], UniMessage], alone: bool = False) -> None:
        self._send(worker, data, alone, False)




from typing import List, Dict, Any

from unipipeline import UniKafkaBroker


class KafkaBroker(UniKafkaBroker):
    def get_boostrap_servers(self) -> List[str]:
        return ['localhost:9092']

    def get_security_conf(self) -> Dict[str, Any]:
        return dict()



from unipipeline.answer.uni_answer_message import UniAnswerMessage
from unipipeline.brokers.uni_amqp_pika_broker import UniAmqpPikaBroker, UniAmqpPikaBrokerMessageManager, UniAmqpPikaBrokerConfig
from unipipeline.brokers.uni_broker import UniBroker
from unipipeline.brokers.uni_broker_consumer import UniBrokerConsumer
from unipipeline.brokers.uni_broker_message_manager import UniBrokerMessageManager
from unipipeline.brokers.uni_kafka_broker import UniKafkaBroker
from unipipeline.brokers.uni_kafka_broker_config import UniKafkaBrokerConfig
from unipipeline.brokers.uni_kafka_broker_message_manager import UniKafkaBrokerMessageManager
from unipipeline.brokers.uni_log_broker import UniLogBroker
from unipipeline.brokers.uni_memory_broker import UniMemoryBroker
from unipipeline.brokers.uni_memory_broker_message_manager import UniMemoryBrokerMessageManager
from unipipeline.config.uni_config import UniConfig
from unipipeline.definitions.uni_broker_definition import UniBrokerDefinition
from unipipeline.definitions.uni_cron_task_definition import UniCronTaskDefinition
from unipipeline.definitions.uni_definition import UniDefinition
from unipipeline.definitions.uni_external_definition import UniExternalDefinition
from unipipeline.definitions.uni_message_definition import UniMessageDefinition
from unipipeline.definitions.uni_module_definition import UniModuleDefinition
from unipipeline.definitions.uni_service_definition import UniServiceDefinition
from unipipeline.definitions.uni_waiting_definition import UniWaitingDefinition
from unipipeline.definitions.uni_worker_definition import UniWorkerDefinition
from unipipeline.errors.uni_answer_delay_error import UniAnswerDelayError
from unipipeline.errors.uni_config_error import UniConfigError
from unipipeline.errors.uni_definition_not_found_error import UniDefinitionNotFoundError
from unipipeline.errors.uni_error import UniError
from unipipeline.errors.uni_payload_error import UniPayloadError
from unipipeline.errors.uni_sending_to_worker_error import UniSendingToWorkerError
from unipipeline.errors.uni_work_flow_error import UniWorkFlowError
from unipipeline.message.uni_cron_message import UniCronMessage
from unipipeline.message.uni_message import UniMessage
from unipipeline.message_meta.uni_message_meta import UniMessageMeta, UniMessageMetaErr, UniMessageMetaErrTopic
from unipipeline.modules.uni import Uni
from unipipeline.modules.uni_cron_job import UniCronJob
from unipipeline.modules.uni_mediator import UniMediator
from unipipeline.utils.complex_serializer import complex_serializer_json_dumps
from unipipeline.utils.connection_pool import ConnectionObj, ConnectionRC, ConnectionManager, ConnectionPool, connection_pool
from unipipeline.utils.uni_util import UniUtil
from unipipeline.waiting.uni_postgres_waiting import UniPostgresWaiting
from unipipeline.waiting.uni_wating import UniWaiting
from unipipeline.worker.uni_worker import UniWorker
from unipipeline.worker.uni_worker_consumer import UniWorkerConsumer
from unipipeline.worker.uni_worker_consumer_manager import UniWorkerConsumerManager
from unipipeline.worker.uni_worker_consumer_message import UniWorkerConsumerMessage

__all__ = (
    "Uni",
    "UniConfig",
    "UniMediator",
    "UniDefinition",
    "UniModuleDefinition",
    "UniServiceDefinition",
    "UniExternalDefinition",
    "UniUtil",
    "UniAnswerMessage",

    "complex_serializer_json_dumps",

    "ConnectionObj",
    "ConnectionRC",
    "ConnectionManager",
    "ConnectionPool",
    "connection_pool",

    # cron
    "UniCronJob",
    "UniCronTaskDefinition",

    # broker
    "UniBrokerMessageManager",
    "UniBroker",
    "UniBrokerConsumer",
    "UniBrokerDefinition",

    "UniAmqpPikaBroker",
    "UniAmqpPikaBrokerConfig",
    "UniAmqpPikaBrokerMessageManager",

    "UniLogBroker",

    "UniMemoryBroker",
    "UniMemoryBrokerMessageManager",

    "UniKafkaBroker",
    "UniKafkaBrokerConfig",
    "UniKafkaBrokerMessageManager",

    # message
    "UniMessage",
    "UniMessageDefinition",
    "UniMessageMeta",
    "UniMessageMetaErr",
    "UniMessageMetaErrTopic",
    "UniCronMessage",

    # worker
    "UniWorker",
    "UniWorkerConsumer",
    "UniWorkerConsumerManager",
    "UniWorkerConsumerMessage",
    "UniWorkerDefinition",

    # waiting
    "UniWaitingDefinition",
    "UniWaiting",
    "UniPostgresWaiting",

    # Error
    "UniError",
    "UniDefinitionNotFoundError",
    "UniConfigError",
    "UniPayloadError",
    "UniSendingToWorkerError",
    "UniWorkFlowError",
    "UniAnswerDelayError",
)



from typing import NamedTuple, Callable

from unipipeline.brokers.uni_broker_message_manager import UniBrokerMessageManager
from unipipeline.message_meta.uni_message_meta import UniMessageMeta


class UniBrokerConsumer(NamedTuple):
    topic: str
    id: str
    group_id: str
    unwrapped: bool
    prefetch_count: int
    message_handler: Callable[[UniMessageMeta, UniBrokerMessageManager], None]



from typing import TypeVar, Generic, Optional, Type, Any, Union, Dict, TYPE_CHECKING

from unipipeline import UniWorkFlowError
from unipipeline.errors.uni_payload_error import UniPayloadParsingError, UniAnswerPayloadParsingError
from unipipeline.errors.uni_sending_to_worker_error import UniSendingToWorkerError
from unipipeline.answer.uni_answer_message import UniAnswerMessage
from unipipeline.brokers.uni_broker_message_manager import UniBrokerMessageManager
from unipipeline.message.uni_message import UniMessage
from unipipeline.message_meta.uni_message_meta import UniMessageMeta, UniMessageMetaErrTopic, UniAnswerParams
from unipipeline.worker.uni_worker import UniWorker
from unipipeline.worker.uni_worker_consumer_manager import UniWorkerConsumerManager
from unipipeline.worker.uni_worker_consumer_message import UniWorkerConsumerMessage
from unipipeline.definitions.uni_worker_definition import UniWorkerDefinition

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator

TInputMsgPayload = TypeVar('TInputMsgPayload', bound=UniMessage)
TAnswerMsgPayload = TypeVar('TAnswerMsgPayload', bound=Optional[UniMessage])


class UniWorkerConsumer(Generic[TInputMsgPayload, TAnswerMsgPayload]):
    def __init__(self, definition: UniWorkerDefinition, mediator: 'UniMediator', worker_type: Type[UniWorker[TInputMsgPayload, TAnswerMsgPayload]]) -> None:
        self._definition = definition
        self._mediator = mediator

        self._worker_manager = UniWorkerConsumerManager(self.send_to)
        self._worker = worker_type(self._worker_manager)
        self._uni_echo = mediator.echo.mk_child(f'worker[{definition.name}]')

        self._input_message_type: Type[TInputMsgPayload] = mediator.get_message_type(self._definition.input_message.name)  # type: ignore
        self._answer_message_type: Optional[Type[TAnswerMsgPayload]] = mediator.get_message_type(self._definition.answer_message.name) if self._definition.answer_message is not None else None  # type: ignore

        self._current_meta: Optional[UniMessageMeta] = None

    def send_to(self, worker: Union[Type['UniWorker[Any, Any]'], str], data: Union[Dict[str, Any], UniMessage], alone: bool = False, need_answer: bool = False) -> Optional[UniAnswerMessage[UniMessage]]:
        wd = self._mediator.config.get_worker_definition(worker)
        if wd.name not in self._definition.output_workers:
            raise UniSendingToWorkerError(f'worker {wd.name} is not defined in workers->{self._definition.name}->output_workers')
        if need_answer and not wd.need_answer:
            raise UniWorkFlowError(f'you will get no response form worker {wd.name}')
        if need_answer:
            answ_params = UniAnswerParams(topic=self._definition.answer_topic, id=self._worker_manager.id)
            return self._mediator.send_to(wd.name, data, self._current_meta, answer_params=answ_params, alone=alone)
        self._mediator.send_to(wd.name, data, self._current_meta, answer_params=None, alone=alone)
        return None

    def process_message(self, meta: UniMessageMeta, manager: UniBrokerMessageManager) -> None:
        self._current_meta = meta
        msg = UniWorkerConsumerMessage[TInputMsgPayload](self._input_message_type, manager, meta)

        try:
            result: Optional[Union[TAnswerMsgPayload, Dict[str, Any]]] = self._worker.handle_message(msg)
        except UniAnswerPayloadParsingError as e:
            self._mediator.move_to_error_topic(self._definition, meta, UniMessageMetaErrTopic.HANDLE_MESSAGE_ERR, e)
        except UniPayloadParsingError as e:
            self._mediator.move_to_error_topic(self._definition, meta, UniMessageMetaErrTopic.MESSAGE_PAYLOAD_ERR, e)
        # except Exception as e:  # TODO: correct error handling
        #     self._mediator.move_to_error_topic(self._definition, meta, UniMessageMetaErrTopic.HANDLE_MESSAGE_ERR, e)
        else:
            if self._definition.need_answer:
                try:
                    self._mediator.answer_to(self._definition.name, meta, result, self._definition.answer_unwrapped)
                except UniSendingToWorkerError:
                    pass

        if self._definition.ack_after_success:
            msg.ack()
        self._current_meta = None



from unipipeline.utils.uni_util_color import UniUtilColor
from unipipeline.utils.uni_util_template import UniUtilTemplate


class UniUtil:
    def __init__(self) -> None:
        self._color = UniUtilColor()
        self._template = UniUtilTemplate()

    @property
    def color(self) -> UniUtilColor:
        return self._color

    @property
    def template(self) -> UniUtilTemplate:
        return self._template



from typing import Dict, Any, Union, Optional

from unipipeline.errors.uni_payload_error import UniPayloadSerializationError
from unipipeline.brokers.uni_broker import UniBroker
from unipipeline.config.uni_config import UniConfig, UniConfigError
from unipipeline.utils.uni_echo import UniEcho
from unipipeline.modules.uni_mediator import UniMediator
from unipipeline.message.uni_message import UniMessage
from unipipeline.utils.uni_util import UniUtil
from unipipeline.waiting.uni_wating import UniWaiting
from unipipeline.worker.uni_worker import UniWorker


def camel_case(snake_cased: str) -> str:
    return ''.join(word.title() for word in snake_cased.split('_'))


class Uni:
    def __init__(self, config: Union[UniConfig, str], echo_level: Optional[Union[str, int]] = None) -> None:
        self._util = UniUtil()
        self._util.template.set_filter('camel', camel_case)

        self._echo = UniEcho('UNI', level=echo_level, colors=self._util.color)

        if isinstance(config, str):
            config = UniConfig(self._util, self._echo, config)
        if not isinstance(config, UniConfig):
            raise ValueError(f'invalid config type. {type(config).__name__} was given')
        self._mediator = UniMediator(self._util, self._echo, config)

    @property
    def echo(self) -> UniEcho:
        return self._mediator.echo

    @property
    def util(self) -> UniUtil:
        return self._util

    @property
    def config(self) -> UniConfig:
        return self._mediator.config

    def set_echo_level(self, level: int) -> None:
        self._mediator.set_echo_level(level)

    def scaffold(self) -> None:
        try:
            for waiting_def in self._mediator.config.waitings.values():
                waiting_def.type.import_class(UniWaiting, self.echo, auto_create=True, create_template_params=waiting_def, util=self._util)

            for broker_def in self._mediator.config.brokers.values():
                broker_def.type.import_class(UniBroker, self.echo, auto_create=True, create_template_params=broker_def, util=self._util)

            for message_def in self._mediator.config.messages.values():
                message_def.type.import_class(UniMessage, self.echo, auto_create=True, create_template_params=message_def, util=self._util)

            for worker_def in self._mediator.config.workers.values():
                if worker_def.marked_as_external:
                    continue
                assert worker_def.type is not None
                worker_def.type.import_class(UniWorker, self.echo, auto_create=True, create_template_params=worker_def, util=self._util)

        except UniConfigError as e:
            self.echo.exit_with_error(str(e))

    def check(self) -> None:
        try:
            for waiting_def in self._mediator.config.waitings.values():
                waiting_def.type.import_class(UniWaiting, self.echo, util=self._util)

            for broker_def in self._mediator.config.brokers.values():
                broker_def.type.import_class(UniBroker, self.echo, util=self._util)

            for message_def in self._mediator.config.messages.values():
                message_def.type.import_class(UniMessage, self.echo, util=self._util)

            for worker_def in self._mediator.config.workers.values():
                if worker_def.marked_as_external:
                    continue
                assert worker_def.type is not None
                worker_def.type.import_class(UniWorker, self.echo, util=self._util)

        except UniConfigError as e:
            self.echo.exit_with_error(str(e))

    def start_cron(self) -> None:
        try:
            self._mediator.start_cron()
        except KeyboardInterrupt:
            self.echo.log_warning('interrupted')
            exit(0)

    def initialize_cron_producer_workers(self) -> None:
        for t in self._mediator.config.cron_tasks.values():
            self.init_producer_worker(t.worker.name)

    def initialize(self, everything: bool = False) -> None:
        if everything:
            for wn in self._mediator.config.workers.keys():
                self._mediator.add_worker_to_init_list(wn, no_related=True)
        self._mediator.initialize(create=True)

    def init_cron(self) -> None:
        for task in self._mediator.config.cron_tasks.values():
            self._mediator.add_worker_to_init_list(task.worker.name, no_related=True)

    def init_producer_worker(self, name: str) -> None:
        self._mediator.add_worker_to_init_list(name, no_related=True)

    def init_consumer_worker(self, name: str) -> None:
        self._mediator.add_worker_to_init_list(name, no_related=False)
        self._mediator.add_worker_to_consume_list(name)

    def send_to(self, name: str, data: Union[Dict[str, Any], UniMessage], alone: bool = False) -> None:
        try:
            self._mediator.send_to(name, data, alone=alone)
        except UniPayloadSerializationError as e:
            self.echo.exit_with_error(f'invalid props in message: {e}')

    def start_consuming(self) -> None:
        try:
            self._mediator.start_consuming()
        except KeyboardInterrupt:
            self.echo.log_warning('interrupted')
            exit(0)



import os.path
import sys

CWD = os.path.dirname(os.path.abspath(__file__))

sys.path.append(os.path.dirname(CWD))

from unipipeline import Uni
from example.args import args

u = Uni(f"{CWD}/dag-{args.type}.yml")

u.initialize_cron_producer_workers()

u.initialize()

u.start_cron()



from example.messages.ender_after_cron_answer_message import EnderAfterCronAnswerMessage
from example.messages.ender_after_cron_input_message import EnderAfterCronInputMessage
from unipipeline import UniWorker, UniWorkerConsumerMessage, UniCronMessage, UniAnswerMessage


class MySuperCronWorker(UniWorker[UniCronMessage, None]):
    def handle_message(self, msg: UniWorkerConsumerMessage[UniCronMessage]) -> None:
        answ: UniAnswerMessage[EnderAfterCronAnswerMessage] = self.manager.get_answer_from('ender_after_cron_worker', EnderAfterCronInputMessage(value=f'cron>>>{msg.payload.task_name}'))

        print(answ.payload.value)  # noqa



from collections import deque
from typing import Callable, Deque, Tuple, Optional, Dict

from unipipeline.utils.uni_echo import UniEcho
from unipipeline.message_meta.uni_message_meta import UniMessageMeta
from unipipeline.brokers.uni_broker_message_manager import UniBrokerMessageManager
from unipipeline.brokers.uni_memory_broker_message_manager import UniMemoryBrokerMessageManager

TConsumer = Callable[[UniMessageMeta, UniBrokerMessageManager], None]


class UniMemoryBrokerQueue:
    def __init__(self, echo: UniEcho) -> None:
        self._echo = echo
        self._waiting_for_process: Deque[Tuple[int, UniMessageMeta]] = deque()
        self._in_process: Optional[Tuple[int, UniMessageMeta]] = None

        self._msg_counter: int = 0
        self._lst_counter: int = 0
        self._listeners: Dict[int, Tuple[TConsumer, int]] = dict()

    def add(self, msg: UniMessageMeta) -> int:
        self._msg_counter += 1
        msg_id = self._msg_counter
        self._waiting_for_process.append((msg_id, msg))
        self._echo.log_info(f'added msg_id={msg_id} :: {msg}')
        return msg_id

    def get_next(self) -> UniMessageMeta:
        id, answ = self.reserve_next()
        self.mark_as_processed(id)
        return answ

    def move_back_from_reserved(self, msg_id: int) -> None:
        self._echo.log_debug(f'move_back_from_reserved msg_id={msg_id}')
        if self._in_process is None:
            return

        (msg_id_, meta) = self._in_process
        if msg_id != msg_id_:
            return

        self._waiting_for_process.appendleft((msg_id, meta))

    def reserve_next(self) -> Tuple[int, UniMessageMeta]:
        if self._in_process is not None:
            return self._in_process

        item = self._waiting_for_process.popleft()
        self._in_process = item
        self._echo.log_debug(f'reserve_next msg_id={item[0]}')
        return item

    def mark_as_processed(self, msg_id: int) -> None:
        self._echo.log_debug(f'mark_as_processed msg_id={msg_id}')
        if self._in_process is None:
            return

        (msg_id_, meta) = self._in_process
        if msg_id != msg_id_:
            return

        self._in_process = None

    def add_listener(self, listener: TConsumer, prefetch: int) -> int:
        lsg_id = self._lst_counter
        self._lst_counter += 1
        self._listeners[lsg_id] = (listener, prefetch)
        return lsg_id

    def rm_listener(self, lst_id: int) -> None:
        if lst_id not in self._listeners:
            return
        self._listeners.pop(lst_id)

    def messages_to_process_count(self) -> int:
        return len(self._waiting_for_process) + (0 if self._in_process is None else 1)

    def has_messages_to_process(self) -> bool:
        return self.messages_to_process_count() > 0

    def process_all(self) -> None:
        self._echo.log_debug(f'process_all len_listeners={len(self._listeners)} :: messages={self.messages_to_process_count()}')
        if len(self._listeners) == 0:
            return

        while self.has_messages_to_process():
            for lst_id in self._listeners.keys():
                if not self.has_messages_to_process():
                    break

                if lst_id not in self._listeners:
                    continue

                (lst, prefetch) = self._listeners[lst_id]

                for i in range(prefetch):
                    if not self.has_messages_to_process():
                        break

                    (msg_id, meta) = self.reserve_next()
                    manager = UniMemoryBrokerMessageManager(self, msg_id)

                    self._echo.log_info(f'process_all :: lsg_id={lst_id} :: i={i} :: msg_id={msg_id} :: {meta}')
                    lst(meta, manager)
                    self._echo.log_debug(f'process_all len_listeners={len(self._listeners)} :: messages={self.messages_to_process_count()}')



import pytest

from unipipeline.utils.uni_util_color import UniUtilColor


@pytest.mark.parametrize('enabled,color,text,expected', [
    (True, UniUtilColor.COLOR_RED, 'some', '\u001b[31msome\u001b[0m'),
    (False, UniUtilColor.COLOR_RED, 'some', 'some'),
])
def test_color(enabled: bool, color: str, text: str, expected: str) -> None:
    uc = UniUtilColor(enabled)
    assert uc.color_it(color, text) == expected



from unipipeline.errors.uni_error import UniError


class UniPayloadError(UniError):
    pass


class UniPayloadParsingError(UniPayloadError):
    pass


class UniAnswerPayloadParsingError(UniPayloadParsingError):
    pass


class UniPayloadSerializationError(UniPayloadError):
    pass



from typing import Set, Optional
from uuid import UUID

from unipipeline.definitions.uni_broker_definition import UniBrokerDefinition
from unipipeline.definitions.uni_definition import UniDefinition
from unipipeline.definitions.uni_message_definition import UniMessageDefinition
from unipipeline.definitions.uni_module_definition import UniModuleDefinition
from unipipeline.definitions.uni_waiting_definition import UniWaitingDefinition


class UniWorkerDefinition(UniDefinition):
    id: UUID
    name: str
    broker: UniBrokerDefinition
    type: Optional[UniModuleDefinition]
    topic: str
    prefetch_count: int
    error_topic: str
    error_payload_topic: str
    answer_topic: str
    input_message: UniMessageDefinition
    answer_message: Optional[UniMessageDefinition]
    output_workers: Set[str]
    ack_after_success: bool
    waitings: Set[UniWaitingDefinition]
    external: Optional[str]
    input_unwrapped: bool
    answer_unwrapped: bool
    answer_avg_delay_s: int

    @property
    def marked_as_external(self) -> bool:
        return self.external is not None

    @property
    def need_answer(self) -> bool:
        return self.answer_message is not None
